# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine artifact views."""

import abc
import os.path
from enum import StrEnum
from functools import cached_property
from typing import Any, assert_never, override

from django.db.models import F, QuerySet
from django.db.models.functions import Lower
from django.http import Http404, HttpRequest, StreamingHttpResponse
from django.http.response import HttpResponseBase
from django.shortcuts import get_object_or_404
from django.template.response import TemplateResponse
from django.utils.http import http_date
from rest_framework import status

from debusine.artifacts.models import ArtifactCategory
from debusine.db.models import Artifact, ArtifactRelation, FileInArtifact
from debusine.server.tar import TarArtifact
from debusine.web.views.base import (
    BaseUIView,
    DetailViewBase,
    SingleObjectMixinBase,
    WorkspaceView,
)
from debusine.web.views.files import (
    FileDownloadMixin,
    FileUI,
    FileWidget,
    PathMixin,
)
from debusine.web.views.http_errors import HttpError400, catch_http_errors
from debusine.web.views.places import Place
from debusine.web.views.view_utils import highlight_as_yaml


class ArtifactView(WorkspaceView, abc.ABC):
    """Base view for UIs showing an artifact or part of it."""

    permission_denied_message = (
        "Non-public artifact: you might need to login "
        "or make a request with a valid Token header"
    )

    #: Artifact for this view
    artifact: Artifact

    @abc.abstractmethod
    def get_artifact(self) -> Artifact:
        """Return the current artifact."""

    @override
    def init_view(self) -> None:
        """Set the current artifact."""
        super().init_view()
        self.artifact = self.get_artifact()
        self.enforce(self.artifact.can_display)


class ArtifactDetailView(ArtifactView, DetailViewBase[Artifact]):
    """Display an artifact and its file(s)."""

    model = Artifact
    pk_url_kwarg = "artifact_id"
    template_name = "web/artifact-detail.html"
    context_object_name = "artifact"

    @override
    def get_artifact(self) -> Artifact:
        return self.get_object()

    @override
    def get_queryset(self) -> QuerySet[Artifact]:
        """Add the select_related we need to the queryset."""
        return super().get_queryset().select_related("workspace__scope")

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @cached_property
    def file_list(self) -> list[FileInArtifact]:
        """Return the files in the artifact."""
        file_list: list[FileInArtifact] = []
        queryset = (
            FileInArtifact.objects.filter(artifact=self.object)
            .select_related("file")
            .order_by(Lower("path"))
        )
        for file_in_artifact in queryset:
            file_list.append(file_in_artifact)
            setattr(
                file_in_artifact,
                "basename",
                os.path.basename(file_in_artifact.path),
            )
        return file_list

    def _current_view_is_specialized(self) -> bool:
        """Tab for the specialized view information will be served."""
        return ArtifactPlugin.plugin_for(self.artifact.category) is not None

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Return context for this view's tabled template."""
        ctx = super().get_context_data(**kwargs)

        plugin_view = ArtifactPlugin.plugin_for(self.artifact.category)

        if plugin_view is not None:
            # Add specific specialized plugin information
            ctx.update(
                **plugin_view(self.artifact).get_context_data(),
                specialized_view=True,
            )

        ctx["artifact_label"] = self.object.get_label()
        ctx["file_list"] = self.file_list
        ctx["artifact_data"] = highlight_as_yaml(self.object.data)

        ctx["forward"] = (
            Artifact.objects.filter(targeted_by__artifact=self.object)
            .annotate(relation_type=F("targeted_by__type"))
            .annotate_complete()
        )
        ctx["reverse_extends"] = (
            Artifact.objects.filter(relations__target=self.object)
            .annotate(relation_type=F("relations__type"))
            .filter(relation_type=ArtifactRelation.Relations.EXTENDS)
            .annotate_complete()
        )
        ctx["relation_count"] = (
            ctx["forward"].count() + ctx["reverse_extends"].count()
        )

        if len(self.file_list) == 1:
            # If there is only one file: display it inline. Otherwise the
            # template will list the files
            ctx["file"] = FileWidget.create(self.file_list[0], file_tag="files")

        ctx["file_count"] = len(self.file_list)

        return ctx

    @override
    def get_template_names(self) -> list[str]:
        """Return the plugin's template_name or the default one."""
        if self._current_view_is_specialized():
            plugin_class = ArtifactPlugin.plugin_for(self.artifact.category)
            assert plugin_class is not None
            return [plugin_class.template_name]

        return [self.template_name]


class FileView(
    PathMixin,
    FileDownloadMixin,
    ArtifactView,
    SingleObjectMixinBase[FileInArtifact],
    abc.ABC,
):
    """Base for all FileInArtifact views."""

    model = FileInArtifact
    object: FileInArtifact

    @override
    def get_queryset(self) -> QuerySet[FileInArtifact]:
        return (
            super()
            .get_queryset()
            .filter(artifact_id=self.kwargs["artifact_id"])
            .select_related("artifact__workspace__scope")
        )

    @override
    def get_object(
        self, queryset: QuerySet[FileInArtifact] | None = None
    ) -> FileInArtifact:
        """Return the FileInArtifact object to show."""
        # Prevent repeated queries: get_artifact calls get_object before
        # dispatch, and Django's BaseDetailView calls get_object in get()
        if not hasattr(self, "object"):
            assert queryset is None
            queryset = self.get_queryset()
            self.object = get_object_or_404(queryset, path=self.kwargs["path"])
        return self.object

    @override
    def get_artifact(self) -> Artifact:
        return self.get_object().artifact

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Add the file to display."""
        ctx = super().get_context_data(**kwargs)
        ctx["file"] = FileWidget.create(self.object)
        return ctx


class FileDetailView(FileView, DetailViewBase[FileInArtifact]):
    """Display a file from an artifact."""

    template_name = "web/fileinartifact-detail.html"
    object: FileInArtifact

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place


class FileDetailViewRaw(FileView):
    """Download a file from an artifact, displayed inline."""

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place_raw

    def get(
        self,
        request: HttpRequest,  # noqa: ARG002, U100
        **kwargs: Any,
    ) -> HttpResponseBase:
        """Download file with disposition inline."""
        self.object = self.get_object()
        ui_info = FileUI.from_file_in_artifact(self.object)
        return self.stream_file(self.object, ui_info, download=False)


class DownloadFormat(StrEnum):
    """Available download formats."""

    AUTO = "auto"
    TAR_GZ = "tar.gz"


class DownloadPathView(PathMixin, FileDownloadMixin, BaseUIView):
    """View to download an artifact (in .tar.gz or list its files)."""

    allow_work_request_tokens = True

    permission_denied_message = (
        "Non-public artifact: you might need to login "
        "or make a request with a valid Token header"
    )

    @override
    @catch_http_errors
    def dispatch(
        self, request: HttpRequest, *args: Any, **kwargs: Any
    ) -> HttpResponseBase:
        """Enforce permissions and setup common members."""
        artifact_id = self.kwargs["artifact_id"]
        try:
            self.artifact = Artifact.objects.get(pk=artifact_id)
        except Artifact.DoesNotExist:
            raise Http404(f"Artifact {artifact_id} does not exist")
        self.set_current_workspace(self.artifact.workspace)
        self.enforce(self.artifact.can_display)
        return super().dispatch(request, *args, **kwargs)

    @cached_property
    def download_format(self) -> DownloadFormat:
        """Return the download format."""
        if (archive := self.request.GET.get("archive", None)) is None:
            return DownloadFormat.AUTO
        try:
            return DownloadFormat(archive)
        except ValueError:
            values = ", ".join(e.value for e in DownloadFormat)
            raise HttpError400(
                f"Invalid archive parameter: {archive!r}. Supported: {values}"
            )

    def get(self, request: HttpRequest, **kwargs: Any) -> HttpResponseBase:
        """Download files from the artifact in .tar.gz."""
        if not self.path:
            # Download the whole artifact
            return self._get_artifact()

        try:
            # Try to return a file
            file_in_artifact = self.artifact.fileinartifact_set.get(
                path=self.path.rstrip("/"), complete=True
            )
            return self._get_file(file_in_artifact)
        except FileInArtifact.DoesNotExist:
            # No file exist
            pass

        # Try to return a .tar.gz / list of files for the directory
        directory_exists = self.artifact.fileinartifact_set.filter(
            path__startswith=self.path, complete=True
        ).exists()
        if directory_exists:
            return self._get_directory()

        # Neither a file nor directory existed, HTTP 404
        ctx = {
            "error": f'Artifact {self.artifact.id} does not have any file '
            f'or directory for "{self.kwargs["path"]}"'
        }
        return TemplateResponse(
            request, "404.html", ctx, status=status.HTTP_404_NOT_FOUND
        )

    def _get_artifact(self) -> HttpResponseBase:
        match self.download_format:
            case DownloadFormat.AUTO:
                files_in_artifact = self.artifact.fileinartifact_set.filter(
                    complete=True
                )
                if files_in_artifact.count() == 1:
                    file_in_artifact = files_in_artifact.first()
                    assert file_in_artifact is not None
                    return self._get_file(file_in_artifact)
                else:
                    return self._get_directory_tar_gz()
            case DownloadFormat.TAR_GZ:
                return self._get_directory_tar_gz()
            case _ as unreachable:
                assert_never(unreachable)

    def _get_directory(self) -> HttpResponseBase:
        match self.download_format:
            case DownloadFormat.TAR_GZ:
                return self._get_directory_tar_gz()
            case _:
                raise HttpError400(
                    "archive argument needed when downloading directories"
                )

    def _get_directory_tar_gz(self) -> HttpResponseBase:
        # Currently due to https://code.djangoproject.com/ticket/33735
        # the .tar.gz file is kept in memory by Django (asgi) and the
        # first byte to be sent to the client happens when the .tar.gz has
        # been all generated. When the Django ticket is fixed the .tar.gz
        # will be served as soon as a file is added and the memory usage will
        # be reduced to TarArtifact._chunk_size_mb

        response = StreamingHttpResponse(
            TarArtifact(
                self.artifact,
                None if self.path == "" else self.path,
            ),
            status=status.HTTP_200_OK,
        )
        response["Content-Type"] = "application/octet-stream"

        directory_name = ""
        if self.path != "":
            directory_name = self.path.removesuffix("/")
            directory_name = directory_name.replace("/", "_")
            directory_name = f"-{directory_name}"

        filename = f"artifact-{self.artifact.id}{directory_name}.tar.gz"
        disposition = f'attachment; filename="{filename}"'
        response["Content-Disposition"] = disposition
        response["Last-Modified"] = http_date(
            self.artifact.created_at.timestamp()
        )

        return response

    def _get_file(self, file_in_artifact: FileInArtifact) -> HttpResponseBase:
        """Return a response that streams the_given file."""
        ui_info = FileUI.from_file_in_artifact(file_in_artifact)
        return self.stream_file(file_in_artifact, ui_info, download=True)


class ArtifactPlugin(abc.ABC):
    """
    Artifacts with specific outputs must subclass it.

    When subclassing, the subclass:
    - Is automatically used by the /artifact/ID/ endpoint
    - Must define "artifact_category" and "template_name"
    - Must implement "get_context_data()"
    """

    model = Artifact
    artifact_category: ArtifactCategory
    template_name: str
    name: str

    _artifact_plugins: dict[str, type["ArtifactPlugin"]] = {}

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:  # noqa: U100
        """Register the plugin."""
        cls._artifact_plugins[cls.artifact_category] = cls

    def __init__(self, artifact: Artifact) -> None:
        """Store the artifact to display."""
        self.artifact = artifact

    @classmethod
    def plugin_for(
        cls, artifact_category: str
    ) -> type["ArtifactPlugin"] | None:
        """Return ArtifactPlugin for artifact_category or None."""
        return cls._artifact_plugins.get(artifact_category)

    @abc.abstractmethod
    def get_context_data(self) -> dict[str, Any]:
        """Return context data for rendering the artifact view."""



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Rename ``system:host_architecture`` to ``system:native_architecture``."""

from django.db import migrations

from debusine.db.migrations._utils import make_data_field_renamer


class Migration(migrations.Migration):
    dependencies = [
        ("db", "0026_rename_publish_target_to_suite"),
    ]

    operations = [
        make_data_field_renamer(
            "Worker",
            "static_metadata",
            {},
            [("system:host_architecture", "system:native_architecture")],
        )
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Scheduler tag support."""

from collections.abc import Collection
from typing import Any

from debusine.utils.tag_utils import (
    CandidateTags,
    DerivationRules,
    ProvenanceRestrictions,
)


class Tagset:
    """Build a set of tags."""

    def __init__(self) -> None:
        """Initialize an empty tag set."""
        # TODO: allowing arbitrary places to provide provenance restrictions
        # for *required* tag sets is equivalent to allowing those places to
        # remove tags, which is likely a security issue

        #: List of provenance restrictions to apply
        self.restrictions: list[ProvenanceRestrictions] = []
        self.derivation_rules: list[DerivationRules] = []
        #: Set of tags being built
        self.tags: set[str] = set()
        #: Log of interventions applied to merge operations
        self.log: list[str] = []

    def add(self, provenance: str, tags: Collection[str] | None) -> None:
        """
        Add tags from a given provenance to the tag set.

        Provenance restrictions are applied, and if a tag or its prefix has a
        provenance allow list that applies, it is enforced.

        Multiple matching allowlists all count, so the tag is only added if its
        provenance is in all relevant allowlists.
        """
        filtered_tags: set[str] = set(tags) if tags else set()
        for restriction in self.restrictions:
            t = restriction.filter_set(provenance, filtered_tags)
            if t != filtered_tags:
                removed = filtered_tags - t
                self.log.append(
                    f"{restriction.name} provenance restrictions"
                    f" forbids tags {" ".join(sorted(removed))}",
                )
            filtered_tags = t
        self.tags.update(filtered_tags)

    def add_candidates(self, candidates: CandidateTags[Any]) -> None:
        """Add tags from the given sets of candidate tags."""
        for provenance, tags in candidates.sets.items():
            self.add(provenance, tags)

    def finalize(self) -> None:
        """Finalize the tag set, applying derivation rules."""
        for rules in self.derivation_rules:
            new_tags = rules.compute(self.tags)
            if new_tags:
                self.log.append(
                    f"{rules.provenance} derivation rules adds tags"
                    f" {", ".join(sorted(new_tags))}"
                )
            self.add(rules.provenance, new_tags)


def make_task_provided_tagset() -> Tagset:
    """Create a new Tagset configured for task-provided tags."""
    import debusine.tasks.tags as ttags

    tagset = Tagset()
    tagset.restrictions.append(ttags.restrictions_when_provided)
    # TODO: add restrictions from django settings?
    # TODO: configure inferences
    return tagset


def make_task_required_tagset() -> Tagset:
    """Create a new Tagset configured for task-required tags."""
    import debusine.worker.tags as wtags

    tagset = Tagset()
    tagset.restrictions.append(wtags.restrictions_when_required)
    return tagset


def make_worker_provided_tagset() -> Tagset:
    """Create a new Tagset configured for worker-provided tags."""
    import debusine.worker.tags as wtags

    tagset = Tagset()
    tagset.restrictions.append(wtags.restrictions_when_provided)
    # TODO: add restrictions from django settings?
    # TODO: configure inferences
    return tagset


def make_worker_required_tagset() -> Tagset:
    """Create a new Tagset configured for worker-required tags."""
    import debusine.tasks.tags as ttags

    tagset = Tagset()
    tagset.restrictions.append(ttags.restrictions_when_required)
    return tagset



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""UI helpers for collection relations."""

from functools import cached_property
from typing import override

from django.urls import reverse
from django.utils.html import format_html

from debusine.db.models.collections import CollectionRelation
from debusine.web.views import ui_shortcuts
from debusine.web.views.places import ResourcePlace
from debusine.web.views.ui.base import UI


class CollectionRelationUI(UI[CollectionRelation]):
    """UI helpers for CollectionRelation instances."""

    @cached_property
    def place(self) -> ResourcePlace:
        """Return a place to list similar relations for this collection."""
        url = reverse(
            "workspaces:collections:collection_relations:list",
            kwargs={
                "wname": self.instance.source.workspace.name,
                "ccat": self.instance.source.category,
                "cname": self.instance.source.name,
            },
        )

        return ResourcePlace(
            title=f"Relations of {self.instance.source} collection",
            url=url,
            breadcrumb="Relations",
            parent=self.instance.source.ui(self.request).place,
        )

    @override
    def build_ui_shortcuts(self) -> list[ui_shortcuts.UIShortcut]:
        result = super().build_ui_shortcuts()

        if self.can_delete:
            result.append(
                ui_shortcuts.create_collection_relation_delete(self.instance)
            )

        return result

    @cached_property
    def place_delete(self) -> ResourcePlace:
        """Return a place to delete a CollectionRelation."""
        relation = self.instance

        title = format_html(
            "Delete relation for <code>{relation}</code>", relation=relation
        )

        return ResourcePlace(
            title=title,
            url=reverse(
                "workspaces:collections:collection_relations:delete",
                kwargs={
                    "wname": relation.source.workspace.name,
                    "ccat": relation.source.category,
                    "cname": relation.source.name,
                    "pk": relation.pk,
                },
            ),
            parent=self.place,
            breadcrumb="Delete",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine command line interface, archive token management commands.

These are "porcelain": they implement opinionated operations on
:collection:`debian:archive` and :collection:`debian:suite` collections,
rather than basic "plumbing" operations.
"""

import abc
import argparse
from collections.abc import Iterable
from typing import override

import rich
from rich.table import Table

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.base import ModelCommand, WorkspaceCommand
from debusine.client.models import (
    CollectionRoles,
    CollectionTokenGrantRequest,
    CollectionTokenGrantResponse,
)


class TokenCommand(
    ModelCommand[CollectionTokenGrantResponse], WorkspaceCommand, abc.ABC
):
    """Common infrastructure for archive token commands."""

    @override
    def __init__(
        self, parser: argparse.ArgumentParser, args: argparse.Namespace
    ) -> None:
        super().__init__(parser, args)
        if args.suite is not None:
            self.collection_category = CollectionCategory.SUITE
            self.collection_name = args.suite
        else:
            self.collection_category = CollectionCategory.ARCHIVE
            self.collection_name = "_"

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument(
            "--suite",
            help="Operate only on this suite, rather than the whole archive",
        )

    @override
    def _list_rich(
        self, grants: Iterable[CollectionTokenGrantResponse]
    ) -> None:
        table = Table(box=rich.box.MINIMAL_DOUBLE_HEAD)
        table.add_column("Token type")
        table.add_column("Token hash")
        table.add_column("Username")
        table.add_column("Comment")
        table.add_column("Role")
        for grant in grants:
            table.add_row(
                grant.token_type,
                grant.token_hash,
                grant.username,
                grant.comment,
                grant.role,
            )
        rich.print(table)

    @override
    def _show_rich(self, grant: CollectionTokenGrantResponse) -> None:
        table = Table(box=rich.box.SIMPLE, show_header=False)
        table.add_row("Token type:", grant.token_type)
        table.add_row("Token hash:", grant.token_hash)
        table.add_row("Token key (only shown once):", grant.token_key)
        table.add_row("Username:", grant.username)
        table.add_row("Comment:", grant.comment)
        table.add_row("Role:", grant.role)
        rich.print(table)


class TokenGrant(TokenCommand, name="grant", group="archive.token"):
    """Grant a role on an archive to a token or an external username."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument("--token", help="Hash of an existing token")
        parser.add_argument(
            "--external-username", help="External username for a new token"
        )
        parser.add_argument(
            "--comment", default="", help="Comment for a new token"
        )
        parser.add_argument(
            "role",
            choices=sorted(str(role) for role in CollectionRoles),
            help="Role to grant",
        )

    @override
    def run(self) -> None:
        request_data = {"comment": self.args.comment, "role": self.args.role}
        if self.args.token is not None:
            request_data["token_hash"] = self.args.token
        if self.args.external_username is not None:
            request_data["username"] = self.args.external_username
        with self._api_call_or_fail():
            grant = self.debusine.collection_token_grant_create(
                workspace=self.workspace,
                collection_category=self.collection_category,
                collection_name=self.collection_name,
                request=CollectionTokenGrantRequest(**request_data),
            )

        self.show(grant)


class TokenList(TokenCommand, name="list", group="archive.token"):
    """List tokens that have been granted roles on an archive."""

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            self.list(
                self.debusine.collection_token_grant_iter(
                    workspace=self.workspace,
                    collection_category=self.collection_category,
                    collection_name=self.collection_name,
                )
            )


class TokenRevoke(TokenCommand, name="revoke", group="archive.token"):
    """Revoke a role on an archive from a token."""

    @override
    @classmethod
    def configure(cls, parser: argparse.ArgumentParser) -> None:
        super().configure(parser)
        parser.add_argument("token", help="Hash of a token to revoke")
        parser.add_argument(
            "role",
            choices=sorted(str(role) for role in CollectionRoles),
            help="Role to revoke",
        )

    @override
    def run(self) -> None:
        with self._api_call_or_fail():
            self.debusine.collection_token_grant_delete(
                workspace=self.workspace,
                collection_category=self.collection_category,
                collection_name=self.collection_name,
                token_hash=self.args.token,
                role=self.args.role,
            )

        self.feedback(
            f"Role '{self.args.role}' revoked from {self.args.token}."
        )



# Generated by Django 4.2.20 on 2025-06-06 11:49

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0011_collectionitem_db_ci_archive_index_path_idx'),
    ]
    replaces = [
        ('db', '0153_alter_collectionitem_created_by_workflow_and_more'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collectionitem',
            name='created_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='workflow_created_%(class)s',
                to='db.workrequest',
            ),
        ),
        migrations.AlterField(
            model_name='collectionitem',
            name='removed_by_workflow',
            field=models.ForeignKey(
                null=True,
                on_delete=django.db.models.deletion.SET_NULL,
                related_name='workflow_removed_%(class)s',
                to='db.workrequest',
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command delete_expired."""

import contextlib
import datetime as dt
import io
from collections.abc import Iterable, Iterator
from typing import override
from unittest import mock

from django.db import IntegrityError
from django.utils import timezone

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    Collection,
    CollectionItem,
    File,
    FileInArtifact,
    FileInStore,
    FileStore,
    FileUpload,
    Group,
    Token,
    WorkRequest,
    WorkflowTemplate,
    default_file_store,
    default_workspace,
)
from debusine.db.playground import scenarios
from debusine.db.tests.utils import RunInParallelTransaction
from debusine.django.management.tests import call_command
from debusine.server.collections import DebianEnvironmentsManager
from debusine.server.management.commands.delete_expired import (
    DeleteExpiredArtifacts,
    DeleteExpiredEphemeralGroups,
    DeleteExpiredTokens,
    DeleteExpiredWorkRequests,
    DeleteExpiredWorkspaces,
    DeleteOperation,
)
from debusine.test.django import TestCase, TransactionTestCase
from debusine.test.test_utils import tomorrow


class DeleteExpiredCommandTests(TransactionTestCase):
    """Tests for delete_expired command."""

    @override
    def setUp(self) -> None:
        super().setUp()
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_2.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_2.save()

        ArtifactRelation.objects.create(
            artifact=artifact_1,
            target=artifact_2,
            type=ArtifactRelation.Relations.BUILT_USING,
        )

    def test_delete_expired(self) -> None:
        """Test delete_expired command delete artifacts."""
        self.assertEqual(Artifact.objects.count(), 2)
        self.assertEqual(ArtifactRelation.objects.count(), 1)

        func_dfh = mock.Mock()
        func_dm = mock.Mock()
        with (
            mock.patch(
                "debusine.db.models.collections"
                ".CollectionItemManager.drop_full_history",
                func_dfh,
            ),
            mock.patch(
                "debusine.db.models.collections"
                ".CollectionItemManager.drop_metadata",
                func_dm,
            ),
            self.freeze_time() as now,
        ):
            stdout, stderr, retval = call_command("delete_expired")
        func_dfh.assert_called_once_with(now)
        func_dm.assert_called_once_with(now)

        self.assertEqual(Artifact.objects.count(), 0)
        self.assertEqual(ArtifactRelation.objects.count(), 0)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(retval, 0)

    def test_delete_expired_dry_run_verbose(self) -> None:
        """Test delete_expired artifacts dry-run: no deletion."""
        artifact_id_1, artifact_id_2 = Artifact.objects.all()

        # Add an artifact with a file
        artifact_id_3, _ = self.playground.create_artifact(
            paths=["README"], create_files=True, expiration_delay=1
        )
        artifact_id_3.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_id_3.save()

        work_request_1 = self.playground.create_workflow()
        work_request_1.expiration_delay = dt.timedelta(days=1)
        work_request_1.created_at = timezone.now() - dt.timedelta(days=365)
        work_request_1.save()
        collection_1 = work_request_1.internal_collection
        assert collection_1 is not None

        create_time = timezone.now() - dt.timedelta(days=30)
        with mock.patch("django.utils.timezone.now", return_value=create_time):
            self.playground.create_workspace(
                name="ws", expiration_delay=dt.timedelta(days=7)
            )

        self.playground.create_user_token(
            expire_at=timezone.now() - dt.timedelta(days=1)
        )
        self.playground.create_worker_activation_token(
            expire_at=timezone.now() - dt.timedelta(days=1)
        )

        num_artifacts = Artifact.objects.count()
        num_artifact_relations = ArtifactRelation.objects.count()
        num_files_in_store = FileInStore.objects.count()
        num_work_requests = WorkRequest.objects.count()
        num_collections = Collection.objects.count()
        num_tokens = Token.objects.count()

        stdout, _, _ = call_command("delete_expired", "--dry-run", verbosity=2)

        self.assertEqual(
            stdout,
            "dry-run mode: no changes will be made\n"
            + _format_deleted_work_requests([work_request_1], [collection_1])
            + _format_deleted_artifacts(
                {artifact_id_1, artifact_id_2, artifact_id_3}
            )
            + "There were no unused ephemeral groups\n"
            + "Deleting 1 expired workspaces\n"
            + "Deleting 1 expired tokens\n"
            + "dry-run mode: no changes were made\n",
        )

        # Artifacts, ArtifactRelation didn't get deleted
        self.assertEqual(Artifact.objects.count(), num_artifacts)
        self.assertEqual(
            ArtifactRelation.objects.count(), num_artifact_relations
        )

        # Files didn't get deleted
        self.assertEqual(FileInStore.objects.count(), num_files_in_store)

        # Work requests and their internal collections didn't get deleted
        self.assertEqual(WorkRequest.objects.count(), num_work_requests)
        self.assertEqual(Collection.objects.count(), num_collections)

        # Tokens didn't get deleted
        self.assertEqual(Token.objects.count(), num_tokens)

    def test_delete_expired_dry_run(self) -> None:
        """Test --dry-run without verbose does not write to stdout."""
        stdout, _, _ = call_command("delete_expired", "--dry-run")

        # dry-run without verbose: no messages are written to stdout.
        self.assertEqual(stdout, "")

    def test_delete_expired_verbose(self) -> None:
        """Test delete_expired verbose: print information."""
        artifact_id_1 = Artifact.objects.all()[0]
        artifact_id_2 = Artifact.objects.all()[1]
        work_request_1 = self.playground.create_workflow()
        work_request_1.created_at = timezone.now() - dt.timedelta(days=365)
        work_request_1.expiration_delay = dt.timedelta(days=1)
        work_request_1.save()
        collection_1 = work_request_1.internal_collection
        assert collection_1 is not None
        collection_1.manager.add_artifact(
            artifact_id_1,
            user=self.playground.get_default_user(),
            name="artifact1",
        )
        collection_1.manager.add_artifact(
            artifact_id_2,
            user=self.playground.get_default_user(),
            name="artifact2",
        )
        create_time = timezone.now() - dt.timedelta(days=30)
        with mock.patch("django.utils.timezone.now", return_value=create_time):
            self.playground.create_workspace(
                name="ws", expiration_delay=dt.timedelta(days=7)
            )
        self.assertEqual(WorkRequest.objects.count(), 1)
        num_collections = Collection.objects.count()
        group = self.playground.create_group(name="test", ephemeral=True)
        token = self.playground.create_user_token(
            expire_at=timezone.now() - dt.timedelta(days=1)
        )
        self.playground.create_worker_activation_token(
            expire_at=timezone.now() - dt.timedelta(days=1)
        )

        stdout, _, _ = call_command("delete_expired", verbosity=2)

        self.assertEqual(
            stdout,
            _format_deleted_work_requests([work_request_1], [collection_1])
            + _format_deleted_artifacts({artifact_id_1, artifact_id_2})
            + "Deleting 1 unused ephemeral groups\n"
            + "Deleting 1 expired workspaces\n"
            + "Deleting 1 expired tokens\n",
        )
        self.assertEqual(Artifact.objects.count(), 0)
        self.assertEqual(ArtifactRelation.objects.count(), 0)
        self.assertEqual(WorkRequest.objects.count(), 0)
        self.assertEqual(Collection.objects.count(), num_collections - 1)
        self.assertQuerySetEqual(Group.objects.filter(pk=group.pk), [])
        self.assertQuerySetEqual(Token.objects.filter(pk=token.pk), [])

    def test_delete_expired_verbose_nothing_deleted(self) -> None:
        """Test delete_expired verbose: print no artifact expired."""
        Artifact.objects.update(created_at=tomorrow())
        stdout, _, _ = call_command("delete_expired", verbosity=2)

        self.assertEqual(
            stdout,
            _format_deleted_work_requests([], [])
            + "There were no expired artifacts\n"
            + "There were no unused ephemeral groups\n"
            + "There were no expired workspaces\n"
            + "There were no expired tokens\n",
        )


class DeleteExpiredArtifactsTests(TestCase):
    """Tests for the DeleteExpiredArtifacts class."""

    playground_memory_file_store = False

    @override
    def setUp(self) -> None:
        super().setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.operation = DeleteOperation(out=self.out, err=self.err)
        self.operation.initial_time = timezone.now()
        self.delete_expired = DeleteExpiredArtifacts(self.operation)

    def test_mark_to_keep_artifact_no_expiration_date(self) -> None:
        """One artifact, not expired: keep it."""
        artifact, _ = self.playground.create_artifact(expiration_delay=0)

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(), {artifact}
        )

    def test_mark_to_keep_artifact_expires_tomorrow(self) -> None:
        """One artifact, not expired (expires tomorrow): keep it."""
        artifact, _ = self.playground.create_artifact(expiration_delay=1)

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(), {artifact}
        )

    def test_mark_to_keep_only_artifact_is_expired(self) -> None:
        """One artifact, expired yesterday. No artifacts are kept."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()

        collection_parent_1 = Collection.objects.create(
            name="collection-1",
            category="collection-1",
            workspace=default_workspace(),
            retains_artifacts=Collection.RetainsArtifacts.NEVER,
        )
        CollectionItem.objects.create(
            name="test-2",
            category="test-2",
            parent_collection=collection_parent_1,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_1,
            created_by_user=self.playground.get_default_user(),
        )

        self.assertQuerySetEqual(self.delete_expired._mark_to_keep(), set())

    def test_mark_to_keep_only_artifact_is_retained(self) -> None:
        """One artifact, expired yesterday. retains_artifacts is set."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()

        collection_parent_1 = Collection.objects.create(
            name="collection-1",
            category="collection-1",
            workspace=default_workspace(),
        )
        CollectionItem.objects.create(
            name="test-2",
            category="test-2",
            parent_collection=collection_parent_1,
            child_type=CollectionItem.Types.ARTIFACT,
            artifact=artifact_1,
            created_by_user=self.playground.get_default_user(),
        )

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(), {artifact_1}
        )

    def test_mark_to_keep_artifacts_retained_by_workflow(self) -> None:
        """RetainsArtifacts.WORKFLOWS retains while the workflow is active."""
        template = WorkflowTemplate.objects.create(
            name="test", workspace=default_workspace(), task_name="noop"
        )
        collections: dict[WorkRequest.Statuses, Collection] = {}
        artifacts: dict[WorkRequest.Statuses, Artifact] = {}
        for status in WorkRequest.Statuses:
            workflow = self.playground.create_workflow(
                template,
                task_data={},
            )
            workflow.status = status
            workflow.save()
            assert workflow.internal_collection is not None
            collections[status] = workflow.internal_collection
            collections[
                status
            ].retains_artifacts = Collection.RetainsArtifacts.WORKFLOW
            collections[status].save()
            artifacts[status], _ = self.playground.create_artifact(
                expiration_delay=1
            )
            artifacts[status].created_at = timezone.now() - dt.timedelta(days=2)
            artifacts[status].save()
            collections[status].manager.add_artifact(
                artifacts[status],
                user=self.playground.get_default_user(),
                name=str(status),
            )

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(),
            {
                artifacts[WorkRequest.Statuses.PENDING],
                artifacts[WorkRequest.Statuses.RUNNING],
                artifacts[WorkRequest.Statuses.BLOCKED],
            },
            ordered=False,
        )

    def test_mark_to_keep_all_artifacts_non_expired_or_targeted(self) -> None:
        """Keep both artifacts: the expired is targeted by the not-expired."""
        artifact_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        artifact_expired.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_expired.save()
        artifact_not_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )

        self.playground.create_artifact_relation(
            artifact_not_expired, artifact_expired
        )

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(),
            {artifact_expired, artifact_not_expired},
            ordered=False,
        )

    def test_mark_to_keep_expired_related_to_non_expired(self) -> None:
        """A relates-to relation from an expired artifact does not keep it."""
        artifact_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        artifact_expired.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_expired.save()
        artifact_not_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        self.playground.create_artifact_relation(
            artifact_expired, artifact_not_expired
        )

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(), {artifact_not_expired}
        )

    def test_mark_to_keep_expired_extends_non_expired(self) -> None:
        """An extends relation from an expired artifact keeps it."""
        artifact_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        artifact_expired.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_expired.save()
        artifact_not_expired, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        self.playground.create_artifact_relation(
            artifact_expired,
            artifact_not_expired,
            relation_type=ArtifactRelation.Relations.EXTENDS,
        )

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(),
            {artifact_expired, artifact_not_expired},
            ordered=False,
        )

    def test_mark_to_keep_no_artifact_all_expired_two(self) -> None:
        """Two artifacts (related), both expired."""
        artifact_expired_1, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        artifact_expired_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_expired_1.save()
        artifact_expired_2, _ = self.playground.create_artifact(
            expiration_delay=1
        )
        artifact_expired_2.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_expired_2.save()

        self.playground.create_artifact_relation(
            artifact_expired_1, artifact_expired_2
        )

        self.assertQuerySetEqual(self.delete_expired._mark_to_keep(), set())

    def test_mark_to_keep_no_artifact_all_expired_three(self) -> None:
        """All artifacts are expired."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_2.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_2.save()
        artifact_3, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_3.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_3.save()

        self.playground.create_artifact_relation(artifact_2, artifact_1)
        self.playground.create_artifact_relation(artifact_3, artifact_2)

        self.assertQuerySetEqual(self.delete_expired._mark_to_keep(), set())

    def test_mark_to_keep_two_artifacts_non_expired_and_targeted(self) -> None:
        """Three artifacts, keep two (non-expired and target of non-expired)."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2, _ = self.playground.create_artifact(expiration_delay=0)
        artifact_3, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_3.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_3.save()

        self.playground.create_artifact_relation(artifact_2, artifact_1)
        self.playground.create_artifact_relation(artifact_3, artifact_2)

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(),
            {artifact_1, artifact_2},
            ordered=False,
        )

    def test_mark_to_keep_isolated(self) -> None:
        """Two expired and one non-expired isolated."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_2.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_2.save()
        artifact_3, _ = self.playground.create_artifact(expiration_delay=0)

        self.playground.create_artifact_relation(artifact_2, artifact_1)
        self.playground.create_artifact_relation(artifact_1, artifact_2)

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(), {artifact_3}
        )

    def test_mark_to_keep_artifacts_circular_dependency(self) -> None:
        """Artifacts are not expired (keep), have a circular dependency."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=0)
        artifact_2, _ = self.playground.create_artifact(expiration_delay=0)

        self.playground.create_artifact_relation(artifact_2, artifact_1)
        self.playground.create_artifact_relation(artifact_1, artifact_2)

        self.assertQuerySetEqual(
            self.delete_expired._mark_to_keep(),
            {artifact_1, artifact_2},
            ordered=False,
        )

    def test_sweep_delete_artifacts(self) -> None:
        """Sweep() delete the artifacts and print progress."""
        artifact_1, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_1.save()
        artifact_2, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_2.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_2.save()

        self.playground.create_artifact_relation(artifact_2, artifact_1)
        self.playground.create_artifact_relation(artifact_1, artifact_2)

        artifact_3, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_3.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_3.save()
        artifact_4, _ = self.playground.create_artifact(expiration_delay=0)

        self.playground.create_artifact_relation(artifact_4, artifact_3)

        artifacts_to_keep = self.delete_expired._mark_to_keep()

        self.operation._verbosity = 2
        self.operation.dry_run = False
        self.delete_expired._sweep(artifacts_to_keep)

        self.assertEqual(Artifact.objects.count(), 2)
        self.assertEqual(
            self.out.getvalue(),
            _format_deleted_artifacts({artifact_1, artifact_2}),
        )

    def test_sweep_delete_artifacts_no_retains(self) -> None:
        """Sweep() delete the artifacts from the collections."""
        artifact, _ = self.playground.create_artifact(
            category=ArtifactCategory.SYSTEM_TARBALL,
            data={
                "codename": "bookworm",
                "architecture": "amd64",
                "variant": "apt",
                "with_dev": True,
            },
            expiration_delay=1,
        )
        artifact.created_at = timezone.now() - dt.timedelta(days=2)
        artifact.save()

        collection = Collection.objects.create(
            name="Debian",
            category=CollectionCategory.ENVIRONMENTS,
            workspace=default_workspace(),
            retains_artifacts=Collection.RetainsArtifacts.NEVER,
        )

        manager = DebianEnvironmentsManager(collection=collection)
        manager.add_artifact(
            artifact,
            user=self.playground.get_default_user(),
            variables={"codename": "trixie"},
        )

        artifacts_to_keep = self.delete_expired._mark_to_keep()

        self.operation._verbosity = 2
        self.operation.dry_run = False
        self.delete_expired._sweep(artifacts_to_keep)

        self.assertEqual(Artifact.objects.count(), 0)
        self.assertEqual(
            self.out.getvalue(),
            _format_deleted_artifacts({artifact}),
        )

    def test_delete_artifact_all_related_files(self) -> None:
        """_delete_artifact() deletes all the artifact's related models."""
        uploaded_file_name = "README"
        uploading_file_name = "uploading"
        artifact, _ = self.playground.create_artifact(
            paths=[uploaded_file_name], create_files=True
        )
        FileUpload.objects.create(
            file_in_artifact=FileInArtifact.objects.create(
                artifact=artifact,
                path=uploading_file_name,
                file=self.playground.create_file(),
                complete=False,
            ),
            path="temp_file_uploading",
        )

        # Both files were added in the artifact; one was added in the store,
        # and the other is uploading
        self.assertEqual(FileInArtifact.objects.count(), 2)
        self.assertEqual(FileInStore.objects.count(), 1)
        self.assertEqual(FileUpload.objects.count(), 1)
        self.assertEqual(File.objects.count(), 2)

        # Retrieve the store_backend for the uploaded file
        uploaded_fileobj = File.objects.get(
            artifact=artifact, fileinartifact__path=uploaded_file_name
        )
        store_backend = FileStore.objects.get(
            files=uploaded_fileobj
        ).get_backend_object()

        # The file exists
        filepath = store_backend.get_local_path(uploaded_fileobj)
        assert filepath is not None
        self.assertTrue(filepath.exists())

        # The Artifact is going to be deleted and its two files with it
        self.operation.dry_run = False
        self.delete_expired._delete_artifact(artifact)

        self.assertFalse(Artifact.objects.filter(id=artifact.id).exists())
        self.assertEqual(FileInArtifact.objects.count(), 0)
        self.assertEqual(FileUpload.objects.count(), 0)

        # FileInStore and File are not deleted: they are deleted after
        # the artifact is deleted in a different transaction (done in
        # DeleteExpiredArtifacts.run() )
        self.assertEqual(FileInStore.objects.count(), 1)
        self.assertEqual(File.objects.count(), 2)

        self.assertTrue(filepath.exists())

    def test_delete_artifact_not_related_files(self) -> None:
        """
        Two files created, _delete_artifact() delete the Artifact.

        Check that only the files that are supposed to be deleted are deleted.
        """
        file_to_keep_name = "README"  # used in two artifacts
        file_to_delete_name = (
            "another-file"  # used only in the deleted artifact
        )

        # Create the artifacts
        artifact_to_delete, _ = self.playground.create_artifact(
            paths=[file_to_keep_name, file_to_delete_name], create_files=True
        )
        artifact_to_keep, _ = self.playground.create_artifact()

        # Get both files
        fias_to_delete = {
            fia.path: fia for fia in artifact_to_delete.fileinartifact_set.all()
        }
        file_to_keep = fias_to_delete[file_to_keep_name].file
        file_to_delete = fias_to_delete[file_to_delete_name].file

        # Add file_to_keep in the artifact that is not being deleted
        # (so the file is kept)
        fia_to_keep = FileInArtifact.objects.create(
            artifact=artifact_to_keep, file=file_to_keep, complete=True
        )

        store_backend = FileInStore.objects.get(
            file=file_to_keep
        ).store.get_backend_object()

        file_to_keep_path = store_backend.get_local_path(file_to_keep)
        assert file_to_keep_path is not None
        file_to_delete_path = store_backend.get_local_path(file_to_delete)
        assert file_to_delete_path is not None

        # Both files exist on disk
        self.assertTrue(file_to_keep_path.exists())
        self.assertTrue(file_to_delete_path.exists())

        self.operation.dry_run = False

        self.delete_expired._delete_artifact(artifact_to_delete)

        # Only the expected FileInArtifact rows are deleted
        for fia in fias_to_delete.values():
            self.assertFalse(FileInArtifact.objects.filter(id=fia.id).exists())
        self.assertTrue(
            FileInArtifact.objects.filter(id=fia_to_keep.id).exists()
        )

        # Both files still exist on disk (they would be deleted by
        # DeleteExpiredArtifacts._delete_files_from_stores)
        self.assertTrue(file_to_keep_path.exists())
        self.assertTrue(file_to_delete_path.exists())

    def test_delete_artifact_dry_run(self) -> None:
        """_delete_artifact() does nothing: running in dry-run."""
        artifact, _ = self.playground.create_artifact(expiration_delay=1)
        artifact.created_at = timezone.now() - dt.timedelta(days=2)
        artifact.save()

        # By default DeleteExpiredArtifacts runs in dry-run
        self.assertTrue(self.operation.dry_run)

        self.delete_expired._delete_artifact(artifact)

        # Artifact was not deleted
        self.assertTrue(Artifact.objects.filter(id=artifact.id).exists())


class DeleteExpiredArtifactsTransactionTests(TransactionTestCase):
    """Tests for DeleteExpiredArtifacts that require transactions."""

    playground_memory_file_store = False

    @override
    def setUp(self) -> None:
        super().setUp()
        self.workspace = self.playground.get_default_workspace()

        self.artifact_1, _ = self.playground.create_artifact(
            expiration_delay=1, paths=["README"], create_files=True
        )
        self.artifact_1.created_at = timezone.now() - dt.timedelta(days=2)
        self.artifact_1.save()
        self.artifact_2, _ = self.playground.create_artifact(expiration_delay=1)
        self.artifact_2.created_at = timezone.now() - dt.timedelta(days=2)
        self.artifact_2.save()

        self.artifact_relation = self.playground.create_artifact_relation(
            self.artifact_1, self.artifact_2
        )

        self.out = io.StringIO()
        self.err = io.StringIO()

        self.operation = DeleteOperation(
            out=self.out, err=self.err, dry_run=False, verbosity=1
        )
        self.delete_expired = DeleteExpiredArtifacts(self.operation)

    def test_delete_files_from_multiple_stores(self) -> None:
        """_delete_files_from_stores deletes files from more than one store."""
        fileobj = self.playground.create_file()

        # Add the file in the default_file_store()
        FileInStore.objects.create(store=default_file_store(), file=fileobj)

        # Add the same file in a secondary store
        secondary_store = FileStore.objects.create(
            name="Test",
            backend=FileStore.BackendChoices.LOCAL,
            configuration={"base_directory": "/"},
        )
        FileInStore.objects.create(store=secondary_store, file=fileobj)

        # _delete_files_from_stores should delete the file from multiple stores
        self.delete_expired._delete_files_from_stores()

        # And the file is gone...
        self.assertFalse(File.objects.filter(id=fileobj.id).exists())

    def test_delete_files_from_stores(self) -> None:
        """_delete_files_from_stores deletes the file."""
        fileobj = self.playground.create_file()
        FileInStore.objects.create(store=default_file_store(), file=fileobj)

        self.delete_expired._delete_files_from_stores()

        self.assertFalse(File.objects.filter(id=fileobj.id).exists())

    def test_delete_files_from_stores_file_was_re_added(self) -> None:
        """
        _delete_files_from_stores does not delete the file.

        The file exist in another Artifact: cannot be deleted.
        """
        fileobj = self.playground.create_file()

        artifact, _ = self.playground.create_artifact()
        FileInArtifact.objects.create(artifact=artifact, file=fileobj)

        self.delete_expired._delete_files_from_stores()

        self.assertTrue(File.objects.filter(id=fileobj.id).exists())

    def test_run_skips_locked(self) -> None:
        """Locked artifacts are skipped."""
        thread = RunInParallelTransaction(
            lambda: list(
                Artifact.objects.filter(
                    id=self.artifact_2.id
                ).select_for_update()
            )
        )
        thread.start_transaction()

        with self.operation:
            self.delete_expired.run()

        thread.stop_transaction()

        self.assertFalse(
            Artifact.objects.filter(id=self.artifact_1.id).exists()
        )
        self.assertTrue(Artifact.objects.filter(id=self.artifact_2.id).exists())
        self.assertFalse(
            ArtifactRelation.objects.filter(
                id=self.artifact_relation.id
            ).exists()
        )
        self.assertEqual(self.err.getvalue(), "")

    def test_run_skips_locked_and_recomputes_expired_artifacts(self) -> None:
        """
        If some expired artifacts are skipped due to locks, the expiry job
        checks for other artifacts that should be kept as a result.
        """
        artifact_3, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_3.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_3.save()
        artifact_relation_2 = self.playground.create_artifact_relation(
            self.artifact_2, artifact_3
        )
        artifact_4, _ = self.playground.create_artifact(expiration_delay=1)
        artifact_4.created_at = timezone.now() - dt.timedelta(days=2)
        artifact_4.save()
        artifact_relation_3 = self.playground.create_artifact_relation(
            artifact_3, artifact_4
        )
        artifact_5, _ = self.playground.create_artifact()

        # We have a chain of artifacts 1 -> 2 -> 3 -> 4.  If 2 is locked,
        # then the chain below it should be kept, although 1 may still be
        # removed.  A non-expired artifact 5 is also still kept.
        thread = RunInParallelTransaction(
            lambda: list(
                Artifact.objects.filter(
                    id=self.artifact_2.id
                ).select_for_update()
            )
        )
        thread.start_transaction()

        with self.operation:
            self.delete_expired.run()

        thread.stop_transaction()

        self.assertFalse(
            Artifact.objects.filter(id=self.artifact_1.id).exists()
        )
        kept_artifact_ids = {
            self.artifact_2.id,
            artifact_3.id,
            artifact_4.id,
            artifact_5.id,
        }
        self.assertCountEqual(
            Artifact.objects.filter(id__in=kept_artifact_ids).values_list(
                "id", flat=True
            ),
            kept_artifact_ids,
        )
        self.assertFalse(
            ArtifactRelation.objects.filter(
                id=self.artifact_relation.id
            ).exists()
        )
        kept_relation_ids = {
            artifact_relation_2.id,
            artifact_relation_3.id,
        }
        self.assertCountEqual(
            ArtifactRelation.objects.filter(
                id__in=kept_relation_ids
            ).values_list("id", flat=True),
            kept_relation_ids,
        )
        self.assertEqual(self.err.getvalue(), "")

    def test_run_blocks_new_relation(self) -> None:
        """
        A parallel transaction cannot add a new relation on an artifact that
        is being expired.
        """
        thread = RunInParallelTransaction(
            lambda: self.playground.create_artifact_relation(
                self.playground.create_artifact()[0], self.artifact_1
            )
        )
        thread.start_transaction()

        with self.operation:
            self.delete_expired.run()

        with self.assertRaisesRegex(
            IntegrityError, "db_artifactrelation_target_id"
        ):
            thread.stop_transaction()

        self.assertFalse(
            Artifact.objects.filter(
                id__in={self.artifact_1.id, self.artifact_2.id}
            ).exists()
        )
        self.assertFalse(
            ArtifactRelation.objects.filter(
                id=self.artifact_relation.id
            ).exists()
        )
        self.assertEqual(self.err.getvalue(), "")

    def test_on_commit_cleanup_delete_files(self) -> None:
        """File is deleted from the store if the transaction is committed."""
        artifact, _ = self.playground.create_artifact(
            paths=["README"], create_files=True, expiration_delay=1
        )
        artifact.created_at = timezone.now() - dt.timedelta(days=2)
        artifact.save()

        message_artifacts_deleted = _format_deleted_artifacts(
            Artifact.objects.all()
        )

        self.assertTrue(Artifact.objects.filter(id=artifact.id).exists())
        file_in_store = FileInStore.objects.latest("id")
        fileobj = file_in_store.file
        store_backend = file_in_store.store.get_backend_object()

        file_path = store_backend.get_local_path(fileobj)
        assert file_path is not None
        self.assertTrue(file_path.exists())

        with self.operation:
            self.operation._verbosity = 2
            self.delete_expired.run()

        self.assertFalse(Artifact.objects.filter(id=artifact.id).exists())

        # The file does not exist anymore
        self.assertFalse(file_path.exists())

        # Expect two files to be deleted: README from each of
        # self.artifact_1 and artifact.
        self.assertEqual(
            self.out.getvalue(),
            message_artifacts_deleted + "Deleting 2 files from the store\n",
        )

    def test_delete_files_from_stores_skips_locked(self) -> None:
        """Locked files are skipped."""
        expired_file = self.artifact_1.files.get()
        expired_file_path = (
            expired_file.fileinstore_set.get()
            .store.get_backend_object()
            .get_local_path(expired_file)
        )
        assert expired_file_path is not None
        self.assertTrue(expired_file_path.exists())

        thread = RunInParallelTransaction(
            lambda: list(
                File.objects.filter(id=expired_file.id).select_for_update()
            )
        )
        thread.start_transaction()

        with self.operation:
            self.delete_expired.run()

        thread.stop_transaction()

        self.assertTrue(File.objects.filter(id=expired_file.id).exists())
        self.assertTrue(expired_file_path.exists())

    def test_delete_files_from_stores_blocks_new_file_in_artifact(self) -> None:
        """
        A parallel transaction cannot add a new reference to a file that is
        being expired.
        """
        expired_file = self.artifact_1.files.get()
        expired_file_path = (
            expired_file.fileinstore_set.get()
            .store.get_backend_object()
            .get_local_path(expired_file)
        )
        assert expired_file_path is not None
        self.assertTrue(expired_file_path.exists())
        # Delete the expired artifacts: we're only interested in the file
        # phase of expiry.
        self.artifact_relation.delete()
        self.artifact_1.fileinartifact_set.all().delete()
        self.artifact_1.delete()
        self.artifact_2.delete()
        artifact_3, _ = self.playground.create_artifact()

        thread = RunInParallelTransaction(
            lambda: FileInArtifact.objects.create(
                artifact=artifact_3, path="test", file=expired_file
            )
        )
        thread.start_transaction()

        with self.operation:
            self.delete_expired.run()

        with self.assertRaisesRegex(
            IntegrityError, "db_fileinartifact_file_id"
        ):
            thread.stop_transaction()

        self.assertFalse(File.objects.filter(id=expired_file.id).exists())
        self.assertFalse(expired_file_path.exists())

    def test_delete_files_from_stores_raise_artifact_is_deleted(self) -> None:
        """Artifact is deleted even if the file store deletion fail."""
        exception = RuntimeError
        patcher = mock.patch.object(
            self.delete_expired,
            "_delete_files_from_stores",
            autospec=True,
        )
        delete_files_mocked = patcher.start()
        delete_files_mocked.side_effect = exception
        self.addCleanup(patcher.stop)

        with self.operation:
            try:
                self.delete_expired.run()
            except exception:
                pass

        # _delete_files_from_stores raised an exception. Assert that the
        # artifact_2 deletion transaction is committed
        self.assertFalse(
            Artifact.objects.filter(id=self.artifact_2.id).exists()
        )


class DeleteExpiredWorkRequestsTests(TransactionTestCase):
    """Tests for the DeleteExpiredWorkRequests class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.operation = DeleteOperation(
            out=self.out, err=self.err, dry_run=False
        )
        self.operation.initial_time = timezone.now()
        self.delete_expired = DeleteExpiredWorkRequests(self.operation)

    @contextlib.contextmanager
    def mock_perform_deletions(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock perform_deletions."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteExpiredWorkRequests.perform_deletions"
        ) as perform_deletions:
            yield perform_deletions

    @contextlib.contextmanager
    def mock_verbose(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock operation.verbose."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteOperation.verbose"
        ) as verbose:
            yield verbose

    def assert_dry_run_does_not_delete(self) -> None:
        """Make sure that, with dry_run, perform_deletions is not called."""
        old_dry_run = self.operation.dry_run
        self.operation.dry_run = True
        try:
            with self.mock_perform_deletions() as perform_deletions:
                self.delete_expired.run()
            perform_deletions.assert_not_called()
        finally:
            self.operation.dry_run = old_dry_run

    def test_no_expired(self) -> None:
        """No expired work requests."""
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertEqual(self.delete_expired.to_delete.work_requests, set())
        self.assertEqual(self.delete_expired.to_delete.collections, set())
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with("There were no expired work requests\n")

    def test_one_expired(self) -> None:
        """One expired work request."""
        wr1 = self.playground.create_worker_task(expired=True)
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertEqual(self.delete_expired.to_delete.work_requests, {wr1})
        self.assertEqual(self.delete_expired.to_delete.collections, set())
        perform_deletions.assert_called()
        verbose.assert_called_once_with(
            "Deleting 1 expired work requests and 0 expired collections\n"
        )
        self.assert_dry_run_does_not_delete()

    def test_multiple_expired(self) -> None:
        """Multiple expired work requests."""
        wr1 = self.playground.create_worker_task(expired=True)
        wr2 = self.playground.create_workflow()
        wr2.created_at = timezone.now() - dt.timedelta(days=365)
        wr2.expiration_delay = dt.timedelta(days=1)
        wr2.save()
        c2 = wr2.internal_collection
        wr3 = self.playground.create_worker_task(parent=wr2, expired=True)
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertEqual(
            self.delete_expired.to_delete.work_requests, {wr1, wr2, wr3}
        )
        self.assertEqual(self.delete_expired.to_delete.collections, {c2})
        perform_deletions.assert_called()
        verbose.assert_called_once_with(
            "Deleting 3 expired work requests and 1 expired collections\n"
        )
        self.assert_dry_run_does_not_delete()


class DeleteExpiredEphemeralGroupsTests(TransactionTestCase):
    """Tests for the DeleteExpiredEphemeralGroups class."""

    scenario = scenarios.DefaultScopeUser()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.operation = DeleteOperation(
            out=self.out, err=self.err, dry_run=False
        )
        self.operation.initial_time = timezone.now()
        self.delete_expired = DeleteExpiredEphemeralGroups(self.operation)

    @contextlib.contextmanager
    def mock_perform_deletions(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock perform_deletions."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteExpiredEphemeralGroups.perform_deletions"
        ) as perform_deletions:
            yield perform_deletions

    @contextlib.contextmanager
    def mock_verbose(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock operation.verbose."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteOperation.verbose"
        ) as verbose:
            yield verbose

    def assert_dry_run_does_not_delete(self) -> None:
        """Make sure that, with dry_run, perform_deletions is not called."""
        old_dry_run = self.operation.dry_run
        self.operation.dry_run = True
        try:
            with self.mock_perform_deletions() as perform_deletions:
                self.delete_expired.run()
            perform_deletions.assert_not_called()
        finally:
            self.operation.dry_run = old_dry_run

    def test_no_unused(self) -> None:
        """No unused ephemeral groups."""
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.groups, [])
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with(
            "There were no unused ephemeral groups\n"
        )

    def test_one_unused(self) -> None:
        """One unused ephemeral group."""
        group = self.playground.create_group(name="test", ephemeral=True)
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.groups, [group])
        perform_deletions.assert_called()
        verbose.assert_called_once_with("Deleting 1 unused ephemeral groups\n")
        self.assert_dry_run_does_not_delete()

    def test_ephemeral_but_used(self) -> None:
        """One ephemeral group in use."""
        group = self.playground.create_group(name="test", ephemeral=True)
        group.assign_role(self.scenario.scope, self.scenario.scope.Roles.OWNER)
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.groups, [])
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with(
            "There were no unused ephemeral groups\n"
        )


class DeleteExpiredWorkspacesTests(TransactionTestCase):
    """Tests for the DeleteExpiredWorkspaces class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.operation = DeleteOperation(
            out=self.out, err=self.err, dry_run=False
        )
        self.operation.initial_time = timezone.now()
        self.delete_expired = DeleteExpiredWorkspaces(self.operation)

    @contextlib.contextmanager
    def mock_perform_deletions(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock perform_deletions."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteExpiredWorkspaces.perform_deletions"
        ) as perform_deletions:
            yield perform_deletions

    @contextlib.contextmanager
    def mock_verbose(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock operation.verbose."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteOperation.verbose"
        ) as verbose:
            yield verbose

    def assert_dry_run_does_not_delete(self) -> None:
        """Make sure that, with dry_run, perform_deletions is not called."""
        old_dry_run = self.operation.dry_run
        self.operation.dry_run = True
        try:
            with self.mock_perform_deletions() as perform_deletions:
                self.delete_expired.run()
            perform_deletions.assert_not_called()
        finally:
            self.operation.dry_run = old_dry_run

    def test_no_expired(self) -> None:
        """No expired workspaces."""
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.to_delete.workspaces, [])
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with("There were no expired workspaces\n")

    def test_one_expired(self) -> None:
        """One expired workspace."""
        create_time = timezone.now() - dt.timedelta(days=30)
        with mock.patch("django.utils.timezone.now", return_value=create_time):
            ws = self.playground.create_workspace(
                name="ws", expiration_delay=dt.timedelta(days=7)
            )

        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(
            self.delete_expired.to_delete.workspaces.all(), [ws]
        )
        perform_deletions.assert_called()
        verbose.assert_called_once_with("Deleting 1 expired workspaces\n")
        self.assert_dry_run_does_not_delete()

    def test_multiple_expired(self) -> None:
        """Multiple expired workspaces."""
        create_time = timezone.now() - dt.timedelta(days=30)
        with mock.patch("django.utils.timezone.now", return_value=create_time):
            ws1 = self.playground.create_workspace(
                name="ws1", expiration_delay=dt.timedelta(days=7)
            )
            ws2 = self.playground.create_workspace(
                name="ws2", expiration_delay=dt.timedelta(days=7)
            )
            ws3 = self.playground.create_workspace(
                name="ws3", expiration_delay=dt.timedelta(days=7)
            )
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(
            self.delete_expired.to_delete.workspaces,
            [ws1, ws2, ws3],
            ordered=False,
        )
        perform_deletions.assert_called()
        verbose.assert_called_once_with("Deleting 3 expired workspaces\n")
        self.assert_dry_run_does_not_delete()


class DeleteExpiredTokensTests(TransactionTestCase):
    """Tests for the DeleteExpiredTokens class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.out = io.StringIO()
        self.err = io.StringIO()
        self.operation = DeleteOperation(
            out=self.out, err=self.err, dry_run=False
        )
        self.operation.initial_time = timezone.now()
        self.delete_expired = DeleteExpiredTokens(self.operation)

    @contextlib.contextmanager
    def mock_perform_deletions(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock perform_deletions."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteExpiredTokens.perform_deletions"
        ) as perform_deletions:
            yield perform_deletions

    @contextlib.contextmanager
    def mock_verbose(self) -> Iterator[mock.MagicMock]:
        """Shortcut to mock operation.verbose."""
        with mock.patch(
            "debusine.server.management.commands.delete_expired"
            ".DeleteOperation.verbose"
        ) as verbose:
            yield verbose

    def assert_dry_run_does_not_delete(self) -> None:
        """Make sure that, with dry_run, perform_deletions is not called."""
        old_dry_run = self.operation.dry_run
        self.operation.dry_run = True
        try:
            with self.mock_perform_deletions() as perform_deletions:
                self.delete_expired.run()
            perform_deletions.assert_not_called()
        finally:
            self.operation.dry_run = old_dry_run

    def test_no_expired(self) -> None:
        """No expired tokens."""
        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.tokens, [])
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with("There were no expired tokens\n")

    def test_one_expired(self) -> None:
        """One unused token."""
        token = self.playground.create_user_token(
            expire_at=timezone.now() - dt.timedelta(seconds=1)
        )
        self.playground.create_worker_activation_token(
            expire_at=timezone.now() - dt.timedelta(seconds=1)
        )

        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.tokens, [token])
        perform_deletions.assert_called()
        verbose.assert_called_once_with("Deleting 1 expired tokens\n")
        self.assert_dry_run_does_not_delete()

    def test_not_expired(self) -> None:
        """Tokens in use."""
        self.playground.create_user_token()
        self.playground.create_worker_token()
        self.playground.create_worker_activation_token(
            expire_at=timezone.now() + dt.timedelta(days=1)
        )

        with (
            self.mock_perform_deletions() as perform_deletions,
            self.mock_verbose() as verbose,
        ):
            self.delete_expired.run()

        self.assertQuerySetEqual(self.delete_expired.tokens, [])
        perform_deletions.assert_not_called()
        verbose.assert_called_once_with("There were no expired tokens\n")


def _format_deleted_work_requests(
    work_requests: list[WorkRequest], collections: list[Collection]
) -> str:
    message = ""

    if work_requests or collections:
        message += (
            f"Deleting {len(work_requests)} expired work requests"
            f" and {len(collections)} expired collections\n"
        )
    else:
        message += "There were no expired work requests\n"

    return message


def _format_deleted_artifacts(artifacts: Iterable[Artifact]) -> str:
    message = ""

    for artifact in sorted(artifacts, key=lambda x: x.id):
        message += f"Deleted artifact {artifact.id}\n"

    return message



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Data models for db scopes."""

import enum
import re
from typing import Any, Self, TYPE_CHECKING, TypeAlias, cast, override

import pgtrigger
from django.core.exceptions import ImproperlyConfigured, ValidationError
from django.db import models
from django.db.models import (
    CheckConstraint,
    Exists,
    F,
    OuterRef,
    Q,
    QuerySet,
    UniqueConstraint,
)

from debusine.db.models import permissions
from debusine.db.models.files import File, FileStore
from debusine.db.models.permissions import (
    Allow,
    Role,
    permission_check,
    permission_filter,
)
from debusine.db.permissioncontext import PermissionContext

if TYPE_CHECKING:
    from django.http import HttpRequest
    from django_stubs_ext.db.models import TypedModelMeta

    from debusine.server.file_backend.interface import FileBackendInterface
    from debusine.web.views.ui.scopes import ScopeUI
else:
    TypedModelMeta = object

#: Scope names reserved for use in toplevel URL path components
RESERVED_SCOPE_NAMES = frozenset(
    (
        "accounts",
        "admin",
        "api",
        "api-auth",
        "artifact",
        "task-status",
        "user",
        "workers",
        "work-request",
        "workspace",
    )
)

#: Regexp matching the structure of scope names
scope_name_regex = re.compile(r"^[A-Za-z][A-Za-z0-9+._-]*$")


def is_valid_scope_name(value: str) -> bool:
    """Check if value is a valid scope name."""
    if value in RESERVED_SCOPE_NAMES:
        return False
    return bool(scope_name_regex.match(value))


def validate_scope_name(value: str) -> None:
    """Validate scope names."""
    if not is_valid_scope_name(value):
        raise ValidationError(
            "%(value)r is not a valid scope name", params={"value": value}
        )


class ScopeRoleBase(permissions.RoleBase):
    """Scope role implementation."""

    implied_by_scope_roles: frozenset["ScopeRoles"]

    @override
    def _setup(self) -> None:
        """Set up implications for a newly constructed role."""
        implied_by_scope_roles: set[ScopeRoles] = {cast(ScopeRoles, self)}
        for i in self.implied_by:
            match i:
                case Role():
                    # Resolve a role passed during class definition into its
                    # enum instance
                    role = self.__class__(i.value)
                    implied_by_scope_roles |= role.implied_by_scope_roles
                case _:
                    raise ImproperlyConfigured(
                        f"Scope roles do not support implications by {type(i)}"
                    )
        self.implied_by_scope_roles = frozenset(implied_by_scope_roles)

    def q(self, pc: PermissionContext) -> Q:
        """Return a Q expression to select scopes with this role."""
        return Q(
            roles__group__in=pc.groups(),
            roles__role__in=self.implied_by_scope_roles,
        )

    def implies(self, role: "ScopeRoles") -> bool:
        """Check if this role implies the given one."""
        return self in role.implied_by_scope_roles


class ScopeRoles(permissions.Roles, ScopeRoleBase, enum.ReprEnum):
    """Available roles for a Scope."""

    OWNER = Role(
        "owner",
        description="Full administrative access",
    )

    # Read-only access to the whole scope.
    OBSERVER = Role(
        "observer",
        implied_by=[OWNER],
        description="Read-only access to all workspaces",
    )

    # "viewer" is reserved for possibly adding private scopes in the future,
    # in which case we may need a distinction between "can view the scope
    # itself" and "can view all workspaces in the scope".


ScopeRoles.setup()


class ScopeQuerySet[A](QuerySet["Scope", A]):
    """Custom QuerySet for Scope."""

    def with_role(self, pc: PermissionContext, role: ScopeRoles) -> Self:
        """Keep only resources where the user has the given role."""
        if not pc.user.is_authenticated:
            return self.none()
        if pc.user.is_system:
            return self
        return self.filter(
            Exists(self.model.objects.filter(role.q(pc), pk=OuterRef("pk")))
        )

    @permission_filter(work_request=Allow.PASS, anonymous=Allow.PASS)
    def can_display(self, pc: PermissionContext) -> Self:  # noqa: ARG002, U100
        """Keep only Scopes that can be displayed."""
        return self

    @permission_filter()
    def can_create_workspace(self, pc: PermissionContext) -> Self:
        """Keep only Scopes where the user can create workspaces."""
        return self.with_role(pc, Scope.Roles.OWNER)


class ScopeManager(models.Manager["Scope"]):
    """Manager for Scope model."""

    def get_roles_model(self) -> type["ScopeRole"]:
        """Get the model used for role assignment."""
        return ScopeRole

    @override
    def get_queryset(self) -> ScopeQuerySet[Any]:
        """Use the custom QuerySet."""
        return ScopeQuerySet(self.model, using=self._db)


class Scope(models.Model):
    """
    Scope model.

    This is used to create different distinct sets of groups and workspaces
    """

    Roles: TypeAlias = ScopeRoles

    name = models.CharField(
        max_length=255,
        unique=True,
        validators=[validate_scope_name],
        help_text="internal name for the scope",
    )
    label = models.CharField(
        max_length=255,
        unique=True,
        help_text="User-visible name for the scope",
    )
    icon = models.CharField(
        max_length=255,
        default="",
        blank=True,
        help_text=(
            "Optional user-visible icon,"
            " resolved via ``{% static %}`` in templates"
        ),
    )
    file_stores = models.ManyToManyField(
        FileStore, related_name="scopes", through="db.FileStoreInScope"
    )

    objects = ScopeManager.from_queryset(ScopeQuerySet)()

    class Meta(TypedModelMeta):
        base_manager_name = "objects"

    @override
    def __str__(self) -> str:
        """Return basic information of Scope."""
        return self.name

    def ui(self, request: "HttpRequest") -> "ScopeUI":
        """Return a UI helper for this instance."""
        from debusine.web.views.ui.scopes import ScopeUI

        return ScopeUI.get(request, self)

    @permission_check(
        "{user} cannot display scope {resource}",
        work_request=Allow.PASS,
        anonymous=Allow.PASS,
    )
    def can_display(self, pc: PermissionContext) -> bool:  # noqa: ARG002, U100
        """Check if the scope can be displayed."""
        return True

    @permission_check(
        "{user} cannot create workspaces in {resource}",
    )
    def can_create_workspace(self, pc: PermissionContext) -> bool:
        """Check if the user can create workspaces in this scope."""
        return self.has_role(pc, Scope.Roles.OWNER)

    def has_role(self, pc: PermissionContext, role: ScopeRoles) -> bool:
        """Check if the user has the given role on this Scope."""
        from debusine.db.context import context

        if not pc.user.is_authenticated:
            return False
        if pc.user.is_system:
            return True
        if context.pc_is(pc) and context.scope == self:
            return any(r.implies(role) for r in context.scope_roles)
        return Scope.objects.with_role(pc, role).filter(pk=self.pk).exists()

    # See https://github.com/typeddjango/django-stubs/issues/1047 for the typing
    def get_group_roles(
        self, pc: PermissionContext
    ) -> QuerySet["ScopeRole", str]:
        """Get the roles of the user on this scope."""
        if not pc.user.is_authenticated:
            return ScopeRole.objects.none().values_list("role", flat=True)
        else:
            return (
                ScopeRole.objects.filter(resource=self, group__in=pc.groups())
                .values_list("role", flat=True)
                .distinct()
            )

    def upload_file_stores(
        self,
        fileobj: File,
        *,
        enforce_soft_limits: bool = False,
        include_write_only: bool = False,
    ) -> QuerySet[FileStore]:
        """
        Find the file stores in this scope where `fileobj` can be uploaded.

        The returned query set is in descending order of upload priority,
        breaking ties by ascending file store IDs.

        :param enforce_soft_limits: Enforce `soft_max_size` policies as well
          as `max_size`.
        :param include_write_only: Uploading a file to a write-only store
          will mean that debusine won't download it from there again, so
          that only makes sense in specialized situations such as populating
          a backup store.  Set this to True to include write-only stores in
          the returned query set.
        """
        file_stores = (
            self.file_stores.annotate_current_total_size()
            .exclude(filestoreinscope__read_only=True)
            .exclude(max_size__lt=F("current_total_size") + fileobj.size)
            .order_by(
                F("filestoreinscope__upload_priority").desc(nulls_last=True),
                "id",
            )
        )
        if enforce_soft_limits:
            # TODO: We should also enforce FileStoreInScope.soft_max_size.
            file_stores = file_stores.exclude(
                soft_max_size__lt=F("current_total_size") + fileobj.size
            )
        if not include_write_only:
            file_stores = file_stores.exclude(filestoreinscope__write_only=True)
        return file_stores

    def upload_file_backend(self, fileobj: File) -> "FileBackendInterface[Any]":
        """
        Find the best file backend for uploading `fileobj`.

        :raises IndexError: if there is no such file backend.
        """
        return self.upload_file_stores(fileobj)[0].get_backend_object()

    def download_file_stores(self, fileobj: File) -> QuerySet[FileStore]:
        """
        Find the file stores in this scope that have `fileobj`, if any.

        The returned query set is in descending order of download priority,
        breaking ties in descending order of upload priority and then by
        ascending file store IDs.
        """
        return (
            self.file_stores.exclude(filestoreinscope__write_only=True)
            .filter(files=fileobj)
            .order_by(
                F("filestoreinscope__download_priority").desc(nulls_last=True),
                F("filestoreinscope__upload_priority").desc(nulls_last=True),
                "id",
            )
        )

    def download_file_backend(
        self, fileobj: File
    ) -> "FileBackendInterface[Any]":
        """
        Find the best file backend for downloading `fileobj`.

        :raises IndexError: if there is no such file backend.
        """
        return self.download_file_stores(fileobj)[0].get_backend_object()


class ScopeRole(models.Model):
    """Role assignments for scopes."""

    Roles: TypeAlias = ScopeRoles

    resource = models.ForeignKey(
        Scope,
        on_delete=models.CASCADE,
        related_name="roles",
    )

    group = models.ForeignKey(
        "Group",
        on_delete=models.CASCADE,
        related_name="scope_roles",
    )

    role = models.CharField(choices=Roles.choices)

    class Meta(TypedModelMeta):
        constraints = [
            UniqueConstraint(
                fields=["resource", "group", "role"],
                name="%(app_label)s_%(class)s_unique_resource_group_role",
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return a description of the role assignment."""
        return f"{self.group}─{self.role}⟶{self.resource}"


class FileStoreInScope(models.Model):
    """Database model used for extra data on Scope/FileStore relations."""

    scope = models.ForeignKey(Scope, on_delete=models.PROTECT)
    file_store = models.ForeignKey(FileStore, on_delete=models.PROTECT)

    _file_store_instance_wide = models.BooleanField(
        default=True,
        editable=False,
        db_column="file_store_instance_wide",
        help_text="Synced from FileStore.instance_wide; do not update directly",
    )

    #: The priority of this store for the purpose of storing new files.
    #: When adding a new file, debusine tries stores whose policies allow
    #: adding new files in descending order of upload priority, counting
    #: null as the lowest.
    upload_priority = models.IntegerField(blank=True, null=True)

    #: The priority of this store for the purpose of serving files to
    #: clients.  When downloading a file, debusine tries stores in
    #: descending order of download priority, counting null as the lowest;
    #: it breaks ties in descending order of upload priority, again counting
    #: null as the lowest.  If there is still a tie, it picks one of the
    #: possibilities arbitrarily.
    download_priority = models.IntegerField(blank=True, null=True)

    #: If True, the storage maintenance job ensures that this store has a
    #: copy of all files in the scope.
    populate = models.BooleanField(default=False)

    #: If True, the storage maintenance job moves all files in this scope to
    #: some other store in the same scope, following the same rules for
    #: finding a target store as for uploads of new files.  It does not move
    #: into a store if that would take its total size over `soft_max_size`
    #: (either for the scope or the file store), and it logs an error if it
    #: cannot find any eligible target store.
    drain = models.BooleanField(default=False)

    #: If this field is set, then constrain `drain` to use the store with
    #: the given name in this scope.
    #
    # TODO: ruff is correct that we shouldn't use null=True on a TextField,
    # but fixing that retroactively is non-trivial.
    drain_to = models.TextField(blank=True, null=True)  # noqa: DJ001

    #: If True, debusine will not add new files to this store.  Use this in
    #: combination with `drain` to prepare for removing the file store.
    read_only = models.BooleanField(default=False)

    #: If True, debusine will not read files from this store.  This is
    #: suitable for provider storage classes that are designed for long-term
    #: archival rather than routine retrieval, such as S3 Glacier Deep
    #: Archive.
    write_only = models.BooleanField(default=False)

    #: An integer specifying the number of bytes that the file store can
    #: hold for this scope (accounting files that are in multiple scopes to
    #: all of the scopes in question).  This limit may be exceeded
    #: temporarily during uploads; the storage maintenance job will move the
    #: least-recently-used files to another file store to get back below the
    #: limit.
    soft_max_size = models.IntegerField(blank=True, null=True)

    class Meta(TypedModelMeta):
        triggers = [
            pgtrigger.Trigger(
                name="db_filestoreinscope_sync_instance_wide",
                operation=pgtrigger.Insert | pgtrigger.Update,
                when=pgtrigger.Before,
                func=" ".join(
                    """
                    NEW.file_store_instance_wide =
                        (SELECT db_filestore.instance_wide
                         FROM db_filestore
                         WHERE db_filestore.id = NEW.file_store_id);
                    RETURN NEW;
                    """.split()
                ),
            )
        ]
        constraints = [
            UniqueConstraint(
                fields=["scope", "file_store"],
                name="%(app_label)s_%(class)s_unique_scope_file_store",
            ),
            UniqueConstraint(
                fields=["file_store"],
                name=(
                    "%(app_label)s_%(class)s_"
                    "unique_file_store_not_instance_wide"
                ),
                condition=Q(_file_store_instance_wide=False),
            ),
            # It does not make sense to request a store to be populated
            # while also requesting it to be either drained or read-only.
            CheckConstraint(
                name="%(app_label)s_%(class)s_consistent_populate",
                check=Q(populate=False) | Q(drain=False, read_only=False),
            ),
        ]

    @override
    def __str__(self) -> str:
        """Return basic information of FileStoreInScope."""
        return f"{self.scope}/{self.file_store.name}"



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
debusine-admin command to manage groups.

Note: to make commands easier to be invoked from Ansible, we take care to make
them idempotent.
"""

import argparse
from collections.abc import Callable
from functools import cached_property
from typing import Any, NoReturn, Self, cast, override

from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.core.management.base import OutputWrapper
from django.db import transaction

from debusine.db.context import context
from debusine.db.models import Group, Scope, User, Workspace
from debusine.db.models.auth import GroupRoles
from debusine.db.models.scopes import ScopeRole
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import (
    GroupMembers,
    Groups,
    get_scope,
    get_workspace,
)


def print_validation_error(error: ValidationError, file: OutputWrapper) -> None:
    """Print a ValidationError to the given file."""
    for field, errors in error.message_dict.items():
        if field == "__all__":
            lead = ""
        else:
            lead = f"{field}: "
        for msg in errors:
            file.write(f"* {lead}{msg}")


class ParsedGroupArgument:
    """Parsed elements of a ``scope[/workspace]/name`` specification."""

    def __init__(
        self, *, scope: str, name: str, workspace: str | None = None
    ) -> None:
        """
        Construct a ParsedGroupArgument.

        :param scope: scope name
        :param name: group name
        :param workspace: workspace name, or None if the group is not workspaced
        """
        self.scope = scope
        self.name = name
        self.workspace = workspace

    @override
    def __str__(self) -> str:
        parts = [self.scope]
        if self.workspace is not None:
            parts.append(self.workspace)
        parts.append(self.name)
        return "/".join(parts)

    @override
    def __repr__(self) -> str:
        return f"ParsedGroupArgument({self})"

    def as_queryset_filter(self) -> dict[str, str | None]:
        """Build a Group queryset filter kwargs."""
        res: dict[str, str | None] = {
            "scope__name": self.scope,
            "name": self.name,
        }
        if self.workspace is None:
            res["workspace"] = None
        else:
            res["workspace__name"] = self.workspace
        return res

    @cached_property
    def scope_instance(self) -> Scope:
        """Resolve to a Scope instance."""
        try:
            return Scope.objects.get(name=self.scope)
        except Scope.DoesNotExist:
            raise CommandError(f"Scope {self.scope!r} not found", returncode=3)

    @cached_property
    def workspace_instance(self) -> Workspace | None:
        """Resolve to a Workspace instance."""
        if self.workspace is None:
            return None
        else:
            try:
                return Workspace.objects.get(
                    scope__name=self.scope, name=self.workspace
                )
            except Workspace.DoesNotExist:
                name = self.scope + "/" + self.workspace
                raise CommandError(
                    f"Workspace {name!r} not found", returncode=3
                )

    @cached_property
    def group_instance(self) -> Group:
        """Resolve to a Group instance."""
        try:
            return Group.objects.get(**self.as_queryset_filter())
        except Group.DoesNotExist:
            raise CommandError(
                f"Group {self.name!r} not found in scope {self.scope!r}"
                f" and workspace {self.workspace!r}",
                returncode=3,
            )

    @classmethod
    def parse(cls, value: str) -> Self:
        """Build a ParsedGroupArgument from a string argument."""
        tokens = value.split("/")
        match tokens:
            case [scope, name]:
                return cls(scope=scope, name=name)
            case [scope, workspace, name]:
                return cls(scope=scope, name=name, workspace=workspace)
            case _:
                raise CommandError(
                    f"argument {value!r} should be in the form"
                    " 'scopename[/workspacename]/groupname'",
                    returncode=3,
                )


class Command(DebusineBaseCommand):
    """Command to manage groups."""

    help = "Manage groups"

    def add_group_argument(
        self,
        parser: argparse.ArgumentParser,
        help: str,  # noqa: A002
    ) -> None:
        """Add a group identifier argument to an argument parser."""
        parser.add_argument(
            "scope_workspace_group",
            metavar="scope[/workspace]/name",
            help=help,
        )

    def parse_group_argument(
        self, options: dict[str, Any]
    ) -> ParsedGroupArgument:
        """Parse a ``scope[/workspace]/name`` argument."""
        return ParsedGroupArgument.parse(options["scope_workspace_group"])

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        subparsers = parser.add_subparsers(dest="action", required=True)

        group_list = subparsers.add_parser(
            "list", help=self.handle_list.__doc__
        )
        group_list.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )
        group_list.add_argument(
            "--no-workspace",
            action="store_true",
            help="Do not list workspace-specific groups",
        )
        group_list.add_argument(
            "scope_workspace",
            metavar="scope[/workspace]",
            help="list groups for this scope or workspace",
        )

        create = subparsers.add_parser(
            "create", help=self.handle_create.__doc__
        )
        self.add_group_argument(
            create, help="scope/workspace/name identifying the new group"
        )

        rename = subparsers.add_parser(
            "rename", help=self.handle_rename.__doc__
        )
        self.add_group_argument(
            rename, help="scope/workspace/name identifying the group to rename"
        )
        rename.add_argument("name", help="New name for the group")

        delete = subparsers.add_parser(
            "delete", help=self.handle_delete.__doc__
        )
        self.add_group_argument(
            delete, help="scope/workspace/name identifying the group to delete"
        )

        members = subparsers.add_parser(
            "members",
            # Override default usage message since there's otherwise no way
            # to persuade argparse to put the positional argument first:
            # https://github.com/python/cpython/issues/105947
            usage="%(prog)s scope/name [options]",
            help=self.handle_members.__doc__,
        )
        members.add_argument(
            '--yaml', action="store_true", help="Machine readable YAML output"
        )
        self.add_group_argument(
            members,
            help=(
                "scope/workspace/name identifying the group "
                "(must come before --add/--remove/--set options)"
            ),
        )
        members_actions = members.add_mutually_exclusive_group()
        members_actions.add_argument(
            "--add", metavar="user", nargs="+", help="add users"
        )
        members_actions.add_argument(
            "--remove", metavar="user", nargs="+", help="remove users"
        )
        members_actions.add_argument(
            "--set",
            metavar="user",
            nargs="*",
            help="set members to listed users",
        )

        members.add_argument(
            "--role",
            choices=["member", "admin"],
            default="member",
            help="Role to assign to added users (default: member)",
        )

    def get_scope_and_workspace(
        self, scope_workspace: str
    ) -> tuple[Scope, Workspace | None]:
        """Lookup a ``scopename[/workspace]`` string."""
        if "/" not in scope_workspace:
            return get_scope(scope_workspace), None
        else:
            workspace = get_workspace(scope_workspace, require_scope=True)
            return workspace.scope, workspace

    def get_users(self, usernames: list[str]) -> list[User]:
        """Resolve a list of usernames into a list of users."""
        users: list[User] = []
        usernames_ok: bool = True
        errors: list[str] = []
        for username in usernames:
            try:
                users.append(User.objects.get(username=username))
            except User.DoesNotExist:
                usernames_ok = False
                errors.append(f"User {username!r} does not exist")
        if not usernames_ok:
            raise CommandError(
                "\n".join(errors),
                returncode=3,
            )
        return users

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Dispatch the requested action."""
        func = cast(
            Callable[..., NoReturn],
            getattr(self, f"handle_{options['action']}", None),
        )
        if func is None:
            raise CommandError(
                f"Action {options['action']!r} not found", returncode=3
            )

        func(*args, **options)

    def handle_list(
        self,
        no_workspace: bool,
        yaml: bool,
        scope_workspace: str,
        **options: Any,
    ) -> NoReturn:
        """List groups in a scope."""
        scope, workspace = self.get_scope_and_workspace(scope_workspace)
        groups = Group.objects.filter(scope=scope)
        if workspace is not None:
            groups = groups.filter(workspace=workspace)
        elif no_workspace:
            groups = groups.filter(workspace=None)
        Groups(yaml).print(groups, self.stdout)
        raise SystemExit(0)

    @context.disable_permission_checks()
    def handle_create(self, **options: Any) -> NoReturn:
        """
        Create a group, specified as scope/name.

        This is idempotent, and it makes sure the named group exists.
        """
        group_args = self.parse_group_argument(options)
        with transaction.atomic():
            if Group.objects.filter(**group_args.as_queryset_filter()).exists():
                raise SystemExit(0)
            group = Group(
                scope=group_args.scope_instance,
                workspace=group_args.workspace_instance,
                name=group_args.name,
            )
            try:
                group.full_clean()
            except ValidationError as exc:
                self.stderr.write("Created group would be invalid:")
                print_validation_error(exc, file=self.stderr)
                raise SystemExit(3)
            group.save()
        raise SystemExit(0)

    @context.disable_permission_checks()
    def handle_rename(self, *, name: str, **options: Any) -> NoReturn:
        """Rename a group."""
        group_args = self.parse_group_argument(options)
        if name == group_args.name:
            raise SystemExit(0)

        group = group_args.group_instance
        group.name = name
        try:
            group.full_clean()
        except ValidationError as exc:
            self.stderr.write("Renamed group would be invalid:")
            print_validation_error(exc, file=self.stderr)
            raise SystemExit(3)

        group.save()

        raise SystemExit(0)

    def handle_delete(self, **options: Any) -> NoReturn:
        """Delete a group."""
        group_args = self.parse_group_argument(options)
        with transaction.atomic():
            if not Group.objects.filter(
                **group_args.as_queryset_filter()
            ).exists():
                raise SystemExit(0)
            group = group_args.group_instance
            ScopeRole.objects.filter(group=group).delete()
            group.delete()

        raise SystemExit(0)

    @context.disable_permission_checks()
    def handle_members(self, **options: Any) -> NoReturn:
        """Display or change group membership."""
        group_args = self.parse_group_argument(options)
        group = group_args.group_instance
        role = GroupRoles[options["role"].upper()]

        # Change membership
        change_requested: bool = False
        with transaction.atomic():
            if (usernames := options["set"]) is not None:
                change_requested = True
                group.users.clear()
                for user in self.get_users(usernames):
                    group.add_user(user, role=role)
            if usernames := options["add"]:
                change_requested = True
                for user in self.get_users(usernames):
                    group.add_user(user, role=role)
            if usernames := options["remove"]:
                change_requested = True
                for user in self.get_users(usernames):
                    group.remove_user(user)

        if not change_requested:
            # List users
            GroupMembers(options["yaml"]).print(
                group.membership.all(), self.stdout
            )
            raise SystemExit(0)
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""
Debusine tasks.

Import this module using for example:
from debusine.tasks import RunCommandTask, TaskConfigError

Debusine tasks such as sbuild become available
"""

import sys
from typing import cast

from debusine.tasks._task import (  # isort: split
    BaseExternalTask,
    BaseTaskWithExecutor,
    ExtraRepositoryMixin,
    FALLBACK_ARCHITECTURE,
    RunCommandTask,
    TaskConfigError,
    analyze_external_worker_tasks,
    ensure_artifact_categories,
    ensure_collection_category,
    get_provided_external_worker_tags,
)

# Sub-tasks need to be imported in order to be available to BaseExternalTask
# (e.g. for BaseExternalTask.is_valid_task_name). They are registered via
# BaseExternalTask.__init_subclass__.
from debusine.tasks.assemble_signed_source import AssembleSignedSource
from debusine.tasks.autopkgtest import Autopkgtest
from debusine.tasks.blhc import Blhc
from debusine.tasks.debdiff import DebDiff
from debusine.tasks.extract_for_signing import ExtractForSigning
from debusine.tasks.lintian import Lintian
from debusine.tasks.makesourcepackageupload import MakeSourcePackageUpload
from debusine.tasks.mergeuploads import MergeUploads
from debusine.tasks.mmdebstrap import MmDebstrap
from debusine.tasks.noop import Noop
from debusine.tasks.piuparts import Piuparts
from debusine.tasks.sbuild import Sbuild
from debusine.tasks.simplesystemimagebuild import SimpleSystemImageBuild

# Import the documentation from where the code lives
__doc__ = cast(str, sys.modules[BaseExternalTask.__module__].__doc__)

__all__ = [
    "FALLBACK_ARCHITECTURE",
    "AssembleSignedSource",
    "Autopkgtest",
    "BaseExternalTask",
    "BaseTaskWithExecutor",
    "Blhc",
    "DebDiff",
    "ExtraRepositoryMixin",
    "ExtractForSigning",
    "Lintian",
    "MakeSourcePackageUpload",
    "MergeUploads",
    "MmDebstrap",
    "Noop",
    "Piuparts",
    "RunCommandTask",
    "Sbuild",
    "SimpleSystemImageBuild",
    "TaskConfigError",
    "analyze_external_worker_tasks",
    "ensure_artifact_categories",
    "ensure_collection_category",
    "get_provided_external_worker_tags",
]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Catalog of known task tags and task tag prefixes."""

from enum import StrEnum

from debusine.utils.tag_utils import CandidateTags, ProvenanceRestrictions


class ProvenanceProvided(StrEnum):
    """Provenances for task-provided task tags."""

    #: Tags from database queries
    SYSTEM = "system"
    #: Tags from user data
    USER = "user"
    #: Tags from task configuration defined in the workspace
    WORKSPACE = "workspace"


class ProvenanceRequired(StrEnum):
    """Provenances for worker-required task tags."""

    #: Tags required by debusine admins in static worker metadata
    ADMIN = "system"


class TaskProvidedTags(CandidateTags[ProvenanceProvided]):
    """Candidate task-provided tags."""


#: Prefix for scope tags
SCOPE_PREFIX = "task:scope:"
#: Prefix for scope tags
WORKSPACE_PREFIX = "task:workspace:"
#: Prefix for user groups
GROUP_PREFIX = "task:group:"
#: Prefix for source packages
SOURCE_PACKAGE_PREFIX = "task:source-package:"

#: Restrictions applied when adding provided tags
restrictions_when_provided = ProvenanceRestrictions("task-provided")
# task:group:* tags are only allowed to come from database queries
restrictions_when_provided.add_prefix(GROUP_PREFIX, [ProvenanceProvided.SYSTEM])
# task:scope:* tags are only allowed to come from database queries
restrictions_when_provided.add_prefix(SCOPE_PREFIX, [ProvenanceProvided.SYSTEM])
# task:source-package:* tags are only allowed to come from database queries
restrictions_when_provided.add_prefix(
    SOURCE_PACKAGE_PREFIX, [ProvenanceProvided.SYSTEM]
)
# task:workspace:* tags are only allowed to come from database queries
restrictions_when_provided.add_prefix(
    WORKSPACE_PREFIX, [ProvenanceProvided.SYSTEM]
)

#: Restrictions applied when adding required tags
# Use empty restrictions, since adding required tags is not a security
# sensitive operation, but removing them is
restrictions_when_required = ProvenanceRestrictions("task-required")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""DebusineAsyncHttpClient: interact with the debusine server."""

import asyncio
import logging
from typing import Any, TYPE_CHECKING

import aiohttp
from aiohttp import ClientSession

from debusine.utils import get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT
from debusine.worker.config import ConfigHandler

if TYPE_CHECKING:
    from aiohttp.client import _RequestContextManager

logger = logging.getLogger(__name__)


class DebusineAsyncHttpClient:
    """HTTP Client to interact with the debusine server."""

    def __init__(self, config: ConfigHandler) -> None:
        """Initialize DebusineAsyncHttpClient."""
        self._config = config

        self._client_session: ClientSession | None = None

    def post(
        self, path: str, *, json: dict[str, Any], token: str | None = None
    ) -> "_RequestContextManager":
        """
        Make a POST request.

        :param path: used to create the URL (from config.api_url + path).
        :param json: body of the request.
        """
        return self._make_http_request('POST', path, json=json, token=token)

    def put(
        self, path: str, *, json: dict[str, Any]
    ) -> "_RequestContextManager":
        """
        Make a PUT request.

        :param path: used to create the URL (from config.api_url + path).
        :param json: body of the request.
        """
        return self._make_http_request('PUT', path, json=json)

    def _make_http_request(
        self,
        method: str,
        path: str,
        token: str | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> "_RequestContextManager":
        """
        Make an HTTP request to the debusine server.

        Include the token if available in the config.
        """
        self._ensure_client_session()
        assert self._client_session is not None

        method_to_func = {
            'PUT': self._client_session.put,
            'POST': self._client_session.post,
            'GET': self._client_session.get,
        }

        if method not in method_to_func:
            allowed_methods = ", ".join(method_to_func.keys())
            raise ValueError(f'Method must be one of: {allowed_methods}')

        if method == 'GET' and 'data' in kwargs:
            raise ValueError('data not allowed in GET requests')

        if not path.startswith('/'):
            raise ValueError('Path must be absolute')

        headers = kwargs.pop('headers', {})

        if token is not None:
            headers['token'] = token
        elif self._config.token is not None:
            headers['token'] = self._config.token

        url = f'{self._config.api_url}{path}'

        logger.debug(
            "HTTP %s to the server: URL: %s args: %s kwargs: %s",
            method,
            url,
            args,
            kwargs,
        )
        return method_to_func[method](url, *args, **kwargs, headers=headers)

    def _ensure_client_session(
        self, loop: asyncio.AbstractEventLoop | None = None
    ) -> None:
        if self._client_session is None:
            # Use configured timeout if available, otherwise fallback to default
            timeout_value = self._config.http_timeout or DEFAULT_REQUEST_TIMEOUT

            # Convert (connect, read) tuple to total timeout for aiohttp
            if isinstance(timeout_value, tuple):
                connect, read = timeout_value
                timeout = aiohttp.ClientTimeout(
                    total=connect + read,
                    sock_connect=connect,
                    sock_read=read,
                )
            else:
                timeout = aiohttp.ClientTimeout(total=timeout_value)

            headers = {"User-Agent": get_user_agent()}
            self._client_session = aiohttp.ClientSession(
                loop=loop, timeout=timeout, headers=headers
            )

    async def close(self) -> None:
        """Close the client session if needed."""
        if self._client_session is not None:
            await self._client_session.close()
